@@ -16,48 +16,53 @@
 """Storage backend for SWIFT"""
 
 import hashlib
-import httplib
 import logging
 import math
 
-from oslo.config import cfg
-import six.moves.urllib.parse as urlparse
+from oslo_config import cfg
+from oslo_utils import excutils
+from oslo_utils import units
+import six
+from six.moves import http_client
+from six.moves import urllib
 import swiftclient
-import urllib
 
 import glance_store
 from glance_store._drivers.swift import utils as sutils
+from glance_store import capabilities
 from glance_store.common import auth
+from glance_store.common import utils as cutils
 from glance_store import driver
 from glance_store import exceptions
 from glance_store import i18n
 from glance_store import location
-from glance_store.openstack.common import excutils
+
 
 _ = i18n._
 LOG = logging.getLogger(__name__)
 _LI = i18n._LI
 
 DEFAULT_CONTAINER = 'glance'
-DEFAULT_LARGE_OBJECT_SIZE = 5 * 1024  # 5GB
+DEFAULT_LARGE_OBJECT_SIZE = 5 * units.Ki  # 5GB
 DEFAULT_LARGE_OBJECT_CHUNK_SIZE = 200  # 200M
-ONE_MB = 1000 * 1024
+ONE_MB = units.k * units.Ki  # Here we used the mixed meaning of MB
 
 _SWIFT_OPTS = [
-    cfg.BoolOpt('swift_enable_snet', default=False,
-                help=_('Whether to use ServiceNET to communicate with the '
-                       'Swift storage servers.')),
-    cfg.StrOpt('swift_store_auth_version', default='2',
-               help=_('Version of the authentication service to use. '
-                      'Valid versions are 2 for keystone and 1 for swauth '
-                      'and rackspace. (deprecated)')),
     cfg.BoolOpt('swift_store_auth_insecure', default=False,
                 help=_('If True, swiftclient won\'t check for a valid SSL '
                        'certificate when authenticating.')),
+    cfg.StrOpt('swift_store_cacert',
+               help=_('A string giving the CA certificate file to use in '
+                      'SSL connections for verifying certs.')),
     cfg.StrOpt('swift_store_region',
                help=_('The region of the swift endpoint to be used for '
                       'single tenant. This setting is only necessary if the '
                       'tenant has multiple swift endpoints.')),
+    cfg.StrOpt('swift_store_endpoint',
+               default=None,
+               help=_('If set, the configured endpoint will be used. If '
+                      'None, the storage url from the auth response will be '
+                      'used.')),
     cfg.StrOpt('swift_store_endpoint_type', default='publicURL',
                help=_('A string giving the endpoint type of the swift '
                       'service to use (publicURL, adminURL or internalURL). '
@@ -70,7 +75,9 @@ _SWIFT_OPTS = [
     cfg.StrOpt('swift_store_container',
                default=DEFAULT_CONTAINER,
                help=_('Container within the account that the account should '
-                      'use for storing images in Swift.')),
+                      'use for storing images in Swift when using single '
+                      'container mode. In multiple container mode, this will '
+                      'be the prefix for all containers.')),
     cfg.IntOpt('swift_store_large_object_size',
                default=DEFAULT_LARGE_OBJECT_SIZE,
                help=_('The size, in MB, that Glance will start chunking image '
@@ -87,6 +94,18 @@ _SWIFT_OPTS = [
                 help=_('If set to True, enables multi-tenant storage '
                        'mode which causes Glance images to be stored in '
                        'tenant specific Swift accounts.')),
+    cfg.IntOpt('swift_store_multiple_containers_seed',
+               default=0,
+               help=_('When set to 0, a single-tenant store will only use one '
+                      'container to store all images. When set to an integer '
+                      'value between 1 and 32, a single-tenant store will use '
+                      'multiple containers to store images, and this value '
+                      'will determine how many containers are created.'
+                      'Used only when swift_store_multi_tenant is disabled. '
+                      'The total number of containers that will be used is '
+                      'equal to 16^N, so if this config option is set to 2, '
+                      'then 16^2=256 containers will be used to store images.'
+                      )),
     cfg.ListOpt('swift_store_admin_tenants', default=[],
                 help=_('A list of tenants that will be granted read/write '
                        'access on all Swift containers created by Glance in '
@@ -101,12 +120,20 @@ _SWIFT_OPTS = [
                       'before the request fails.'))
 ]
 
-CONF = cfg.CONF
-
-SWIFT_STORE_REF_PARAMS = sutils.SwiftParams().params
-
 
 def swift_retry_iter(resp_iter, length, store, location, context):
+    if not length and isinstance(resp_iter, six.BytesIO):
+        if six.PY3:
+            # On Python 3, io.BytesIO does not have a len attribute, instead
+            # go the end using seek to get the size of the file
+            pos = resp_iter.tell()
+            resp_iter.seek(0, 2)
+            length = resp_iter.tell()
+            resp_iter.seek(pos)
+        else:
+            # On Python 2, StringIO has a len attribute
+            length = resp_iter.len
+
     length = length if length else (resp_iter.len
                                     if hasattr(resp_iter, 'len') else 0)
     retries = 0
@@ -118,7 +145,8 @@ def swift_retry_iter(resp_iter, length, store, location, context):
                 yield chunk
                 bytes_read += len(chunk)
         except swiftclient.ClientException as e:
-            LOG.warn(_(u"Swift exception raised %s") % unicode(e))
+            LOG.warn(_("Swift exception raised %s") %
+                     cutils.exception_to_str(e))
 
         if bytes_read != length:
             if retries == store.conf.glance_store.swift_store_retry_get_count:
@@ -137,9 +165,9 @@ def swift_retry_iter(resp_iter, length, store, location, context):
                           'max_retries': retry_count,
                           'start': bytes_read,
                           'end': length})
-                (resp_headers, resp_iter) = store._get_object(location, None,
-                                                              bytes_read,
-                                                              context=context)
+                (_resp_headers, resp_iter) = store._get_object(location, None,
+                                                               bytes_read,
+                                                               context=context)
         else:
             break
 
@@ -174,7 +202,8 @@ class StoreLocation(location.StoreLocation):
 
     def _get_credstring(self):
         if self.user and self.key:
-            return '%s:%s' % (urllib.quote(self.user), urllib.quote(self.key))
+            return '%s:%s' % (urllib.parse.quote(self.user),
+                              urllib.parse.quote(self.key))
         return ''
 
     def get_uri(self, credentials_included=True):
@@ -190,13 +219,13 @@ class StoreLocation(location.StoreLocation):
         obj = self.obj.strip('/')
 
         if not credentials_included:
-            #Used only in case of an add
-            #Get the current store from config
-            store = CONF.glance_store.default_swift_reference
+            # Used only in case of an add
+            # Get the current store from config
+            store = self.conf.glance_store.default_swift_reference
 
             return '%s://%s/%s/%s' % ('swift+config', store, container, obj)
         if self.scheme == 'swift+config':
-            if self.ssl_enabled == True:
+            if self.ssl_enabled:
                 self.scheme = 'swift+https'
             else:
                 self.scheme = 'swift+http'
@@ -207,9 +236,10 @@ class StoreLocation(location.StoreLocation):
 
     def _get_conf_value_from_account_ref(self, netloc):
         try:
-            self.user = SWIFT_STORE_REF_PARAMS[netloc]['user']
-            self.key = SWIFT_STORE_REF_PARAMS[netloc]['key']
-            netloc = SWIFT_STORE_REF_PARAMS[netloc]['auth_address']
+            ref_params = sutils.SwiftParams(self.conf).params
+            self.user = ref_params[netloc]['user']
+            self.key = ref_params[netloc]['key']
+            netloc = ref_params[netloc]['auth_address']
             self.ssl_enabled = True
             if netloc != '':
                 if netloc.startswith('http://'):
@@ -248,12 +278,12 @@ class StoreLocation(location.StoreLocation):
                 raise exceptions.BadStoreUri(message=reason)
             key = cred_parts.pop()
             user = ':'.join(cred_parts)
-            creds = urllib.unquote(creds)
+            creds = urllib.parse.unquote(creds)
             try:
                 self.user, self.key = creds.rsplit(':', 1)
             except exceptions.BadStoreConfiguration:
-                self.user = urllib.unquote(user)
-                self.key = urllib.unquote(key)
+                self.user = urllib.parse.unquote(user)
+                self.key = urllib.parse.unquote(key)
         else:
             self.user = None
             self.key = None
@@ -297,7 +327,7 @@ class StoreLocation(location.StoreLocation):
             LOG.info(_LI("Invalid store URI: %(reason)s"), {'reason': reason})
             raise exceptions.BadStoreUri(message=reason)
 
-        pieces = urlparse.urlparse(uri)
+        pieces = urllib.parse.urlparse(uri)
         assert pieces.scheme in ('swift', 'swift+http', 'swift+https',
                                  'swift+config')
 
@@ -306,7 +336,7 @@ class StoreLocation(location.StoreLocation):
         path = pieces.path.lstrip('/')
 
         # NOTE(Sridevi): Fix to map the account reference to the
-        # corresponding CONF value
+        # corresponding configuration value
         if self.scheme == 'swift+config':
             netloc = self._get_conf_value_from_account_ref(netloc)
         else:
@@ -327,7 +357,7 @@ class StoreLocation(location.StoreLocation):
             return self.auth_or_store_url
         else:
             if self.scheme == 'swift+config':
-                if self.ssl_enabled == True:
+                if self.ssl_enabled:
                     self.scheme = 'swift+https'
                 else:
                     self.scheme = 'swift+http'
@@ -341,7 +371,8 @@ class StoreLocation(location.StoreLocation):
 
 def Store(conf):
     try:
-        conf.register_opts(_SWIFT_OPTS, group='glance_store')
+        conf.register_opts(_SWIFT_OPTS + sutils.swift_opts,
+                           group='glance_store')
     except cfg.DuplicateOptError:
         pass
 
@@ -349,18 +380,27 @@ def Store(conf):
         return MultiTenantStore(conf)
     return SingleTenantStore(conf)
 
-Store.OPTIONS = _SWIFT_OPTS
+Store.OPTIONS = _SWIFT_OPTS + sutils.swift_opts
+
+
+def _is_slo(slo_header):
+    if (slo_header is not None and isinstance(slo_header, six.string_types)
+            and slo_header.lower() == 'true'):
+        return True
+
+    return False
 
 
 class BaseStore(driver.Store):
 
+    _CAPABILITIES = capabilities.BitMasks.RW_ACCESS
     CHUNKSIZE = 65536
-    OPTIONS = _SWIFT_OPTS
+    OPTIONS = _SWIFT_OPTS + sutils.swift_opts
 
     def get_schemes(self):
         return ('swift+https', 'swift', 'swift+http', 'swift+config')
 
-    def configure(self):
+    def configure(self, re_raise_bsc=False):
         glance_conf = self.conf.glance_store
         _obj_size = self._option_get('swift_store_large_object_size')
         self.large_object_size = _obj_size * ONE_MB
@@ -369,11 +409,12 @@ class BaseStore(driver.Store):
         self.admin_tenants = glance_conf.swift_store_admin_tenants
         self.region = glance_conf.swift_store_region
         self.service_type = glance_conf.swift_store_service_type
+        self.conf_endpoint = glance_conf.swift_store_endpoint
         self.endpoint_type = glance_conf.swift_store_endpoint_type
-        self.snet = glance_conf.swift_enable_snet
         self.insecure = glance_conf.swift_store_auth_insecure
         self.ssl_compression = glance_conf.swift_store_ssl_compression
-        super(BaseStore, self).configure()
+        self.cacert = glance_conf.swift_store_cacert
+        super(BaseStore, self).configure(re_raise_bsc=re_raise_bsc)
 
     def _get_object(self, location, connection=None, start=None, context=None):
         if not connection:
@@ -385,10 +426,10 @@ class BaseStore(driver.Store):
 
         try:
             resp_headers, resp_body = connection.get_object(
-                container=location.container, obj=location.obj,
+                location.container, location.obj,
                 resp_chunk_size=self.CHUNKSIZE, headers=headers)
         except swiftclient.ClientException as e:
-            if e.http_status == httplib.NOT_FOUND:
+            if e.http_status == http_client.NOT_FOUND:
                 msg = _("Swift could not find object %s.") % location.obj
                 LOG.warn(msg)
                 raise exceptions.NotFound(message=msg)
@@ -397,13 +438,9 @@ class BaseStore(driver.Store):
 
         return (resp_headers, resp_body)
 
-    def validate_location(self, uri):
-        pieces = urlparse.urlparse(uri)
-        if pieces.scheme in ['swift+config']:
-            reason = (_("Location credentials are invalid"))
-            raise exceptions.BadStoreUri(message=reason)
-
-    def get(self, location, connection=None, offset=0, chunk_size=None, context=None):
+    @capabilities.check
+    def get(self, location, connection=None,
+            offset=0, chunk_size=None, context=None):
         location = location.store_location
         (resp_headers, resp_body) = self._get_object(location, connection,
                                                      context=context)
@@ -411,7 +448,7 @@ class BaseStore(driver.Store):
         class ResponseIndexable(glance_store.Indexable):
             def another(self):
                 try:
-                    return self.wrapped.next()
+                    return next(self.wrapped)
                 except StopIteration:
                     return ''
 
@@ -427,7 +464,7 @@ class BaseStore(driver.Store):
             connection = self.get_connection(location, context=context)
         try:
             resp_headers = connection.head_object(
-                container=location.container, obj=location.obj)
+                location.container, location.obj)
             return int(resp_headers.get('content-length', 0))
         except Exception:
             return 0
@@ -453,6 +490,7 @@ class BaseStore(driver.Store):
                 LOG.exception(msg % {'container': container,
                                      'chunk': chunk})
 
+    @capabilities.check
     def add(self, image_id, image_file, image_size,
             connection=None, context=None):
         location = self.create_location(image_id, context=context)
@@ -547,7 +585,7 @@ class BaseStore(driver.Store):
                 # Now we write the object manifest and return the
                 # manifest's etag...
                 manifest = "%s/%s-" % (location.container, location.obj)
-                headers = {'ETag': hashlib.md5("").hexdigest(),
+                headers = {'ETag': hashlib.md5(b"").hexdigest(),
                            'X-Object-Manifest': manifest}
 
                 # The ETag returned for the manifest is actually the
@@ -564,7 +602,7 @@ class BaseStore(driver.Store):
             # image data. We *really* should consider NOT returning
             # the location attribute from GET /images/<ID> and
             # GET /images/details
-            if sutils.is_multiple_swift_store_accounts_enabled():
+            if sutils.is_multiple_swift_store_accounts_enabled(self.conf):
                 include_creds = False
             else:
                 include_creds = True
@@ -572,15 +610,16 @@ class BaseStore(driver.Store):
             return (location.get_uri(credentials_included=include_creds),
                     image_size, obj_etag, {})
         except swiftclient.ClientException as e:
-            if e.http_status == httplib.CONFLICT:
+            if e.http_status == http_client.CONFLICT:
                 msg = _("Swift already has an image at this location")
                 raise exceptions.Duplicate(message=msg)
 
             msg = (_(u"Failed to add object to Swift.\n"
-                     "Got error from Swift: %s") % unicode(e))
+                     "Got error from Swift: %s.") % cutils.exception_to_str(e))
             LOG.error(msg)
             raise glance_store.BackendException(msg)
 
+    @capabilities.check
     def delete(self, location, connection=None, context=None):
         location = location.store_location
         if not connection:
@@ -591,17 +630,27 @@ class BaseStore(driver.Store):
             # that means the object was uploaded in chunks/segments,
             # and we need to delete all the chunks as well as the
             # manifest.
-            manifest = None
+            dlo_manifest = None
+            slo_manifest = None
             try:
                 headers = connection.head_object(
                     location.container, location.obj)
-                manifest = headers.get('x-object-manifest')
+                dlo_manifest = headers.get('x-object-manifest')
+                slo_manifest = headers.get('x-static-large-object')
             except swiftclient.ClientException as e:
-                if e.http_status != httplib.NOT_FOUND:
+                if e.http_status != http_client.NOT_FOUND:
                     raise
-            if manifest:
+
+            if _is_slo(slo_manifest):
+                # Delete the manifest as well as the segments
+                query_string = 'multipart-manifest=delete'
+                connection.delete_object(location.container, location.obj,
+                                         query_string=query_string)
+                return
+
+            if dlo_manifest:
                 # Delete all the chunks before the object manifest itself
-                obj_container, obj_prefix = manifest.split('/', 1)
+                obj_container, obj_prefix = dlo_manifest.split('/', 1)
                 segments = connection.get_container(
                     obj_container, prefix=obj_prefix)[1]
                 for segment in segments:
@@ -609,14 +658,19 @@ class BaseStore(driver.Store):
                     # since we're simply sending off parallelizable requests
                     # to Swift to delete stuff. It's not like we're going to
                     # be hogging up network or file I/O here...
-                    connection.delete_object(obj_container,
-                                             segment['name'])
+                    try:
+                        connection.delete_object(obj_container,
+                                                 segment['name'])
+                    except swiftclient.ClientException as e:
+                        msg = _('Unable to delete segment %(segment_name)s')
+                        msg = msg % {'segment_name': segment['name']}
+                        LOG.exception(msg)
 
             # Delete object (or, in segmented case, the manifest)
             connection.delete_object(location.container, location.obj)
 
         except swiftclient.ClientException as e:
-            if e.http_status == httplib.NOT_FOUND:
+            if e.http_status == http_client.NOT_FOUND:
                 msg = _("Swift could not find image at URI.")
                 raise exceptions.NotFound(message=msg)
             else:
@@ -633,7 +687,7 @@ class BaseStore(driver.Store):
         try:
             connection.head_container(container)
         except swiftclient.ClientException as e:
-            if e.http_status == httplib.NOT_FOUND:
+            if e.http_status == http_client.NOT_FOUND:
                 if self.conf.glance_store.swift_store_create_container_on_put:
                     try:
                         msg = (_LI("Creating swift container %(container)s") %
@@ -642,12 +696,13 @@ class BaseStore(driver.Store):
                         connection.put_container(container)
                     except swiftclient.ClientException as e:
                         msg = (_("Failed to add container to Swift.\n"
-                                 "Got error from Swift: %(e)s") % {'e': e})
+                                 "Got error from Swift: %s.") %
+                               cutils.exception_to_str(e))
                         raise glance_store.BackendException(msg)
                 else:
                     msg = (_("The container %(container)s does not exist in "
                              "Swift. Please set the "
-                             "swift_store_create_container_on_put option"
+                             "swift_store_create_container_on_put option "
                              "to add container to Swift automatically.") %
                            {'container': container})
                     raise glance_store.BackendException(msg)
@@ -664,14 +719,23 @@ class BaseStore(driver.Store):
 class SingleTenantStore(BaseStore):
     EXAMPLE_URL = "swift://<USER>:<KEY>@<AUTH_ADDRESS>/<CONTAINER>/<FILE>"
 
-    def configure(self):
-        super(SingleTenantStore, self).configure()
+    def __init__(self, conf):
+        super(SingleTenantStore, self).__init__(conf)
+        self.ref_params = sutils.SwiftParams(self.conf).params
+
+    def configure(self, re_raise_bsc=False):
+        # set configuration before super so configure_add can override
         self.auth_version = self._option_get('swift_store_auth_version')
+        self.user_domain_id = None
+        self.user_domain_name = None
+        self.project_domain_id = None
+        self.project_domain_name = None
+
+        super(SingleTenantStore, self).configure(re_raise_bsc=re_raise_bsc)
 
     def configure_add(self):
-        default_swift_reference = \
-            SWIFT_STORE_REF_PARAMS.get(
-                self.conf.glance_store.default_swift_reference)
+        default_ref = self.conf.glance_store.default_swift_reference
+        default_swift_reference = self.ref_params.get(default_ref)
         if default_swift_reference:
             self.auth_address = default_swift_reference.get('auth_address')
         if (not default_swift_reference) or (not self.auth_address):
@@ -684,8 +748,15 @@ class SingleTenantStore(BaseStore):
         else:
             self.scheme = 'swift+https'
         self.container = self.conf.glance_store.swift_store_container
+        self.auth_version = default_swift_reference.get('auth_version')
         self.user = default_swift_reference.get('user')
         self.key = default_swift_reference.get('key')
+        self.user_domain_id = default_swift_reference.get('user_domain_id')
+        self.user_domain_name = default_swift_reference.get('user_domain_name')
+        self.project_domain_id = default_swift_reference.get(
+            'project_domain_id')
+        self.project_domain_name = default_swift_reference.get(
+            'project_domain_name')
 
         if not (self.user or self.key):
             reason = _("A value for swift_store_ref_params is required.")
@@ -694,13 +765,50 @@ class SingleTenantStore(BaseStore):
                                                    reason=reason)
 
     def create_location(self, image_id, context=None):
+        container_name = self.get_container_name(image_id, self.container)
         specs = {'scheme': self.scheme,
-                 'container': self.container,
+                 'container': container_name,
                  'obj': str(image_id),
                  'auth_or_store_url': self.auth_address,
                  'user': self.user,
                  'key': self.key}
-        return StoreLocation(specs)
+        return StoreLocation(specs, self.conf)
+
+    def get_container_name(self, image_id, default_image_container):
+        """
+        Returns appropriate container name depending upon value of
+        ``swift_store_multiple_containers_seed``. In single-container mode,
+        which is a seed value of 0, simply returns default_image_container.
+        In multiple-container mode, returns default_image_container as the
+        prefix plus a suffix determined by the multiple container seed
+
+        examples:
+            single-container mode:  'glance'
+            multiple-container mode: 'glance_3a1' for image uuid 3A1xxxxxxx...
+
+        :param image_id: UUID of image
+        :param default_image_container: container name from
+               ``swift_store_container``
+        """
+        seed_num_chars = \
+            self.conf.glance_store.swift_store_multiple_containers_seed
+        if seed_num_chars is None \
+                or seed_num_chars < 0 or seed_num_chars > 32:
+            reason = _("An integer value between 0 and 32 is required for"
+                       " swift_store_multiple_containers_seed.")
+            LOG.error(reason)
+            raise exceptions.BadStoreConfiguration(store_name="swift",
+                                                   reason=reason)
+        elif seed_num_chars > 0:
+            image_id = str(image_id).lower()
+
+            num_dashes = image_id[:seed_num_chars].count('-')
+            num_chars = seed_num_chars + num_dashes
+            name_suffix = image_id[:num_chars]
+            new_container_name = default_image_container + '_' + name_suffix
+            return new_container_name
+        else:
+            return default_image_container
 
     def get_connection(self, location, context=None):
         if not location.user:
@@ -712,7 +820,7 @@ class SingleTenantStore(BaseStore):
         if not auth_url.endswith('/'):
             auth_url += '/'
 
-        if self.auth_version == '2':
+        if self.auth_version in ('2', '3'):
             try:
                 tenant_name, user = location.user.split(':')
             except ValueError:
@@ -729,12 +837,20 @@ class SingleTenantStore(BaseStore):
             os_options['region_name'] = self.region
         os_options['endpoint_type'] = self.endpoint_type
         os_options['service_type'] = self.service_type
+        if self.user_domain_id:
+            os_options['user_domain_id'] = self.user_domain_id
+        if self.user_domain_name:
+            os_options['user_domain_name'] = self.user_domain_name
+        if self.project_domain_id:
+            os_options['project_domain_id'] = self.project_domain_id
+        if self.project_domain_name:
+            os_options['project_domain_name'] = self.project_domain_name
 
         return swiftclient.Connection(
-            auth_url, user, location.key, insecure=self.insecure,
-            tenant_name=tenant_name, snet=self.snet,
+            auth_url, user, location.key, preauthurl=self.conf_endpoint,
+            insecure=self.insecure, tenant_name=tenant_name,
             auth_version=self.auth_version, os_options=os_options,
-            ssl_compression=self.ssl_compression)
+            ssl_compression=self.ssl_compression, cacert=self.cacert)
 
 
 class MultiTenantStore(BaseStore):
@@ -751,9 +867,12 @@ class MultiTenantStore(BaseStore):
                        "a service catalog.")
             raise exceptions.BadStoreConfiguration(store_name="swift",
                                                    reason=reason)
-        self.storage_url = auth.get_endpoint(
-            context.service_catalog, service_type=self.service_type,
-            endpoint_region=self.region, endpoint_type=self.endpoint_type)
+        self.storage_url = self.conf_endpoint
+        if not self.storage_url:
+            self.storage_url = auth.get_endpoint(
+                context.service_catalog, service_type=self.service_type,
+                endpoint_region=self.region, endpoint_type=self.endpoint_type)
+
         if self.storage_url.startswith('http://'):
             self.scheme = 'swift+http'
         else:
@@ -781,7 +900,7 @@ class MultiTenantStore(BaseStore):
 
         headers = {}
         if public:
-            headers['X-Container-Read'] = ".r:*,.rlistings"
+            headers['X-Container-Read'] = "*:*"
         elif read_tenants:
             headers['X-Container-Read'] = ','.join('%s:*' % i
                                                    for i in read_tenants)
@@ -798,7 +917,7 @@ class MultiTenantStore(BaseStore):
         try:
             connection.post_container(location.container, headers=headers)
         except swiftclient.ClientException as e:
-            if e.http_status == httplib.NOT_FOUND:
+            if e.http_status == http_client.NOT_FOUND:
                 msg = _("Swift could not find image at URI.")
                 raise exceptions.NotFound(message=msg)
             else:
@@ -810,7 +929,7 @@ class MultiTenantStore(BaseStore):
                  'container': self.container + '_' + str(image_id),
                  'obj': str(image_id),
                  'auth_or_store_url': ep}
-        return StoreLocation(specs)
+        return StoreLocation(specs, self.conf)
 
     def get_connection(self, location, context=None):
         return swiftclient.Connection(
@@ -818,8 +937,9 @@ class MultiTenantStore(BaseStore):
             preauthurl=location.swift_url,
             preauthtoken=context.auth_token,
             tenant_name=context.tenant,
-            auth_version='2', snet=self.snet, insecure=self.insecure,
-            ssl_compression=self.ssl_compression)
+            auth_version='2', insecure=self.insecure,
+            ssl_compression=self.ssl_compression,
+            cacert=self.cacert)
 
 
 class ChunkReader(object):
